@@ -1,6 +1,7 @@
 /* sdlsurface.c --- SDL Surface functions
  *
- * Copyright (C) 2003, 2004, 2005, 2007, 2008, 2009, 2011 Thien-Thi Nguyen
+ * Copyright (C) 2003, 2004, 2005, 2007, 2008, 2009,
+ *   2011, 2013 Thien-Thi Nguyen
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -21,12 +22,21 @@
 #include "guile-sdl.h"
 #include <stdio.h>
 #include <SDL/SDL_image.h>
+#include "snuggle/finangle.h"
+#include "b-uv.h"
 
+#if ! GI_LEVEL_1_8
+IMPORT_SRFI4 ();
+SELECT_MODULE_VAR (mk_u8v,  srfi4, "make-u8vector");
+SELECT_MODULE_VAR (mk_u16v, srfi4, "make-u16vector");
+SELECT_MODULE_VAR (mk_u32v, srfi4, "make-u32vector");
+#define scm_make_u8vector(len,fill)   CALL2 (mk_u8v,  len, fill)
+#define scm_make_u16vector(len,fill)  CALL2 (mk_u16v, len, fill)
+#define scm_make_u32vector(len,fill)  CALL2 (mk_u32v, len, fill)
+#endif
+
+static SCM kp_alpha;
 static SCM alpha_enums;
-static valaka_t alpha_eback[] = {
-  VALAKA (SDL_ALPHA_OPAQUE),
-  VALAKA (SDL_ALPHA_TRANSPARENT)
-};
 
 
 /* smob functions */
@@ -73,11 +83,12 @@ PRIMPROC
   SCM flags),
  doc: /***********
 Return a new surface of dimensions @var{width} by @var{height}.
-Optional third arg @var{flags} (see @code{flagstash:video})
+Optional third arg @var{flags} (@pxref{video flags})
 further specifies the surface.  Color depth and masks
 are those for the current video surface.  */)
 {
 #define FUNC_NAME s_make_surface
+  DECLINIT_SYM2NUM_CC (3, btw->video_flags);
   Uint32 cflags;
   const SDL_PixelFormat *fmt;
   const SDL_Surface *cur;
@@ -89,7 +100,7 @@ are those for the current video surface.  */)
   fmt = cur ? cur->format : SDL_GetVideoInfo ()->vfmt;
   cflags = (UNBOUNDP (flags)
             ? (cur ? cur->flags : 0)
-            : GSDL_FLAGS2ULONG (flags, btw->video_flags, 3));
+            : FLAGS2ULONG (3, flags));
 
   /* Return a newly allocated surface smob.  */
   RETURN_NEW_SURFACE
@@ -113,12 +124,14 @@ PRIMPROC
 Return an empty surface.
 The eight arguments, directly analagous to those
 for SDL_CreateRGBSurface, are: @var{flags}
-(list of symbols, see @code{flagstash:video}),
+(list of symbols, @pxref{video flags}),
 @var{width}, @var{height}, @var{depth}, @var{rmask},
 @var{gmask}, @var{bmask}, @var{amask}
 (all numbers).  */)
 {
 #define FUNC_NAME s_create_rgb_surface
+
+  DECLINIT_SYM2NUM_CC (1, btw->video_flags);
   Uint32 cflags;
 
   ASSERT_INTEGER (width,  2);
@@ -129,7 +142,7 @@ for SDL_CreateRGBSurface, are: @var{flags}
   ASSERT_INTEGER (bmask,  7);
   ASSERT_INTEGER (amask,  8);
 
-  cflags = GSDL_FLAGS2ULONG (flags, btw->video_flags, 1);
+  cflags = FLAGS2ULONG (1, flags);
 
   /* Return a newly allocated surface smob.  */
   RETURN_NEW_SURFACE
@@ -219,19 +232,20 @@ Unlock previously locked @var{surface}.  */)
 
 PRIMPROC
 (load_bmp, "load-bmp", 1, 0, 0,
- (SCM file),
+ (SCM filename),
  doc: /***********
-Return a surface made by loading the bitmap @var{file}.  */)
+Load bitmap data from @var{filename}.
+Return a new surface if successful, otherwise @code{#f}.  */)
 {
 #define FUNC_NAME s_load_bmp
-  range_t cfile;
+  range_t cfilename;
   SDL_Surface *rv;
 
-  ASSERT_STRING (file, 1);
+  ASSERT_STRING (filename, 1);
 
-  FINANGLE (file);
-  rv = SDL_LoadBMP (RS (file));
-  UNFINANGLE (file);
+  FINANGLE (filename);
+  rv = SDL_LoadBMP (RS (filename));
+  UNFINANGLE (filename);
   RETURN_NEW_SURFACE (rv);
 #undef FUNC_NAME
 }
@@ -240,42 +254,20 @@ Return a surface made by loading the bitmap @var{file}.  */)
 /* Load an image in one of many formats.  */
 PRIMPROC
 (load_image, "load-image", 1, 0, 0,
- (SCM file),
+ (SCM filename),
  doc: /***********
-Return a surface made by loading the image @var{file}.
-If there are problems, return @code{#f}.  */)
+Load image data from @var{filename}.
+Return a new surface if successful, otherwise @code{#f}.  */)
 {
 #define FUNC_NAME s_load_image
-  range_t cfile;
+  range_t cfilename;
   SDL_Surface *rv;
 
-  ASSERT_STRING (file, 1);
+  ASSERT_STRING (filename, 1);
 
-  FINANGLE (file);
-  rv = IMG_Load (RS (file));
-  UNFINANGLE (file);
-  RETURN_NEW_SURFACE (rv);
-#undef FUNC_NAME
-}
-
-
-/* Experimental: Load an image in one of many formats from a string.  */
-PRIMPROC
-(string_to_image, "string->image", 1, 0, 0,
- (SCM s),
- doc: /***********
-Return a surface made by loading image data from string
-@var{s}.  [WARNING: This procedure is experimental!]  */)
-{
-#define FUNC_NAME s_string_to_image
-  range_t cs;
-  SDL_Surface *rv;
-
-  ASSERT_STRING (s, 1);
-
-  FINANGLE_RAW (s);
-  rv = IMG_Load_RW (SDL_RWFromConstMem (RS (s), RLEN (s)), 0);
-  UNFINANGLE (s);
+  FINANGLE (filename);
+  rv = IMG_Load (RS (filename));
+  UNFINANGLE (filename);
   RETURN_NEW_SURFACE (rv);
 #undef FUNC_NAME
 }
@@ -284,42 +276,82 @@ Return a surface made by loading image data from string
 PRIMPROC
 (save_bmp, "save-bmp", 2, 0, 0,
  (SCM surface,
-  SCM file),
+  SCM filename),
  doc: /***********
-Save @var{surface} to @var{file} in Windows BMP format.
+Save @var{surface} to @var{filename} in Windows BMP format.
 Return @code{#t} if successful.  */)
 {
 #define FUNC_NAME s_save_bmp
-  range_t cfile;
+  range_t cfilename;
   int rv;
 
   ASSERT_SURFACE (surface,  1);
-  ASSERT_STRING (file, 2);
+  ASSERT_STRING (filename, 2);
 
-  FINANGLE (file);
-  rv = SDL_SaveBMP (UNPACK_SURFACE (surface), RS (file));
-  UNFINANGLE (file);
+  FINANGLE (filename);
+  rv = SDL_SaveBMP (UNPACK_SURFACE (surface), RS (filename));
+  UNFINANGLE (filename);
   RETURN_TRUE_IF_0 (rv);
 #undef FUNC_NAME
 }
 
 
+PRIMPROC
+(surface_color_key_x, "surface-color-key!", 2, 1, 0,
+ (SCM surface, SCM pixel, SCM rle),
+ doc: /***********
+Set the color key for @var{surface} to @var{pixel}.
+If @var{pixel} is @code{#f}, clear the current color key.
+Otherwise, it should be an integer of the appropriate depth
+for @var{surface} (e.g., in the range [0,65535] for 16 bpp).
+If color key processing is enabled, optional arg @var{rle} is a
+boolean that enables (true) or disables (false, the default)
+RLE acceleration.
+Return @code{#t} if successful.  */)
+{
+#define FUNC_NAME s_surface_color_key_x
+  Uint32 cpixel;
+  Uint32 cflags;
+
+  ASSERT_SURFACE (surface, 1);
+  if (EXACTLY_FALSEP (pixel))
+    cflags = cpixel = 0;                /* disable */
+  else
+    {
+      ASSERT_ULONG_COPY (pixel, 2);
+      UNBOUND_MEANS_FALSE (rle);
+      cflags = SDL_SRCCOLORKEY | (NOT_FALSEP (rle)
+                                  ? SDL_RLEACCEL
+                                  : 0);
+    }
+
+  RETURN_TRUE_IF_0
+    (SDL_SetColorKey (UNPACK_SURFACE (surface),
+                      cflags, cpixel));
+#undef FUNC_NAME
+}
+
+
 PRIMPROC
 (set_color_key, "set-color-key!", 3, 0, 0,
  (SCM surface,
   SCM flag,
   SCM key),
  doc: /***********
+NB: This procedure is obsoleted by @code{surface-color-key!}
+and @strong{will be removed} after 2013-12-31.
+
 Set @var{surface} color key as specified by @var{flag}
-(see @code{flagstash:video}) and @var{key}.  */)
+(@pxref{video flags}) and @var{key}.  */)
 {
 #define FUNC_NAME s_set_color_key
+  DECLINIT_SYM2NUM_CC (2, btw->video_flags);
   Uint32 cflag;
 
   ASSERT_SURFACE (surface, 1);
   ASSERT_INTEGER (key, 3);
 
-  cflag = GSDL_FLAGS2ULONG (flag, btw->video_flags, 2);
+  cflag = FLAGS2ULONG (2, flag);
 
   RETURN_TRUE_IF_0
     (SDL_SetColorKey (UNPACK_SURFACE (surface), cflag, C_LONG (key)));
@@ -327,19 +359,60 @@ Set @var{surface} color key as specified by @var{flag}
 }
 
 
+PRIMPROC
+(surface_alpha_x, "surface-alpha!", 2, 1, 0,
+ (SCM surface, SCM alpha, SCM rle),
+ doc: /***********
+Set alpha blending for the entire @var{surface} to @var{alpha}.
+If @var{alpha} is @code{#f}, disable alpha blending.
+Otherwise it should be an integer in the range [0,255]
+or one of the symbols @code{transparent} or @code{opaque}.
+If alpha blending is enabled, optional arg @var{rle} is a
+boolean that enables (true) or disables (false, the default)
+RLE acceleration.
+Return @code{#t} if successful.  */)
+{
+#define FUNC_NAME s_surface_alpha_x
+  DECLINIT_SYM2NUM_CC (2, kp_alpha);
+  Uint32 cflags;
+  Uint8 calpha;
+
+  ASSERT_SURFACE (surface, 1);
+  if (EXACTLY_FALSEP (alpha))
+    cflags = calpha = 0;                /* disable */
+  else
+    {
+      UNBOUND_MEANS_FALSE (rle);
+      cflags = SDL_SRCALPHA | (NOT_FALSEP (rle)
+                               ? SDL_RLEACCEL
+                               : 0);
+      calpha = ENUM2LONG (2, alpha);
+    }
+
+  RETURN_TRUE_IF_0
+    (SDL_SetAlpha (UNPACK_SURFACE (surface),
+                   cflags, calpha));
+#undef FUNC_NAME
+}
+
+
 PRIMPROC
 (set_alpha, "set-alpha!", 2, 1, 0,
  (SCM surface,
   SCM flag,
   SCM alpha),
  doc: /***********
+NB: This procedure is obsoleted by @code{surface-alpha!}
+and @strong{will be removed} after 2013-12-31.
+
 Adjust whole-@var{surface} alpha as specified by
-@var{flag} (see @code{flagstash:video}) and @var{alpha}
-(one of the @code{alpha-enums}, or a number 0-255).
-@xref{Enums and Constants}.
+@var{flag} (@pxref{video flags}) and @var{alpha}
+(@pxref{alpha-enum enums}, or a number 0-255).
 If @var{flag} is @code{#f}, ignore @var{alpha} completely.  */)
 {
 #define FUNC_NAME s_set_alpha
+  DECLINIT_SYM2NUM_CC (2, btw->video_flags);
+  DECLINIT_SYM2NUM_CC (3, alpha_enums);
   Uint32 cflag;
   Uint8 calpha;
 
@@ -356,8 +429,8 @@ If @var{flag} is @code{#f}, ignore @var{alpha} completely.  */)
 
   cflag = (EXACTLY_FALSEP (flag)
            ? 0
-           : GSDL_FLAGS2ULONG (flag, btw->video_flags, 2));
-  calpha = (Uint8) GSDL_ENUM2LONG (alpha, alpha_enums, 3);
+           : FLAGS2ULONG (2, flag));
+  calpha = (Uint8) ENUM2LONG (3, alpha);
 
   RETURN_TRUE_IF_0
     (SDL_SetAlpha (UNPACK_SURFACE (surface),
@@ -420,16 +493,17 @@ PRIMPROC
  doc: /***********
 Convert @var{surface} to the same @var{format} as another
 surface.  Optional third arg @var{flags} is a list of flags
-(symbols) from the set returned by @code{flagstash:video}.  */)
+(@pxref{video flags}).  */)
 {
 #define FUNC_NAME s_convert_surface
+  DECLINIT_SYM2NUM_CC (3, btw->video_flags);
   Uint32 cflags = 0;
 
   ASSERT_SURFACE (surface, 1);
   ASSERT_PIXEL_FORMAT (format, 2);
 
   if (BOUNDP (flags))
-    cflags = GSDL_FLAGS2ULONG (flags, btw->video_flags, 3);
+    cflags = FLAGS2ULONG (3, flags);
 
   RETURN_NEW_SURFACE
     (SDL_ConvertSurface (UNPACK_SURFACE (surface),
@@ -508,6 +582,11 @@ to x=0, y=0, @var{dst} surface dimensions.  */)
 
 /* flipping */
 
+#define MAKE_DEST_SURFACE(src)                          \
+  SDL_CreateRGBSurface (src->flags, src->w, src->h,     \
+                        src->format->BitsPerPixel,      \
+                        0, 0, 0, 0)
+
 PRIMPROC
 (vertical_flip_surface, "vertical-flip-surface", 1, 0, 0,
  (SCM surface),
@@ -528,7 +607,7 @@ Return a new surface created by flipping @var{surface} vertically.  */)
   h = src->h;
 
   /* Create a new surface.  */
-  dst = SDL_CreateRGBSurface (src->flags, w, h, 16, 0, 0, 0, 0);
+  dst = MAKE_DEST_SURFACE (src);
 
   /* Initialize the rects.  */
   srcrect.x = 0;  srcrect.y = 0;    srcrect.w = w;  srcrect.h = 1;
@@ -568,7 +647,7 @@ Return a new surface created by flipping @var{surface} horizontally.  */)
   h = src->h;
 
   /* Create a new surface.  */
-  dst = SDL_CreateRGBSurface (src->flags, w, h, 16, 0, 0, 0, 0);
+  dst = MAKE_DEST_SURFACE (src);
 
   /* Initialize the rects.  */
   srcrect.x = 0;    srcrect.y = 0;  srcrect.w = 1;  srcrect.h = h;
@@ -602,7 +681,92 @@ both vertically and horizontally.  */)
 }
 
 
+PRIMPROC
+(surface_pixels, "surface-pixels", 1, 1, 0,
+ (SCM surface, SCM squash),
+ doc: /***********
+Return pixel data of @var{surface} as a new uniform vector.
+The uvec has type @code{u8}, @code{u16} or @code{u32}, corresponding
+to the @var{surface} depth, with @var{height} x @var{width} elements.
+A 24bpp surface --- @var{depth-in-bytes} of 3 --- is expanded (per pixel)
+to @code{u32}, leaving the high nybble clear.
+
+Optional arg @var{squash} non-@code{#f} means to
+return a u8vector regardless of @var{surface} depth,
+with @var{height} x @var{width} x @var{depth-in-bytes} elements.  */)
+{
+#define FUNC_NAME s_surface_pixels
+  DECLARE_UV_STRUCT (/* not const */, u8,  Uint8);
+  DECLARE_UV_STRUCT (/* not const */, u16, Uint16);
+  DECLARE_UV_STRUCT (/* not const */, u32, Uint32);
+
+  SDL_Surface *src;
+  size_t len, bypp, sz;
+  SCM rv = SCM_EOL;
+
+  ASSERT_SURFACE (surface, 1);
+  src = UNPACK_SURFACE (surface);
+  len = src->w * src->h;
+  bypp = src->format->BytesPerPixel;
+  sz = len * bypp;
+  if (! (1 <= bypp && bypp <= 4))
+    SCM_MISC_ERROR ("invalid depth: ~S", LIST1 (surface));
+
+#define PREP(TT,COUNT)                          \
+  STUFF (TT, rv);                               \
+  rv = scm_make_ ## TT ## vector                \
+    (NUM_ULONG (COUNT), SCM_UNDEFINED);         \
+  GET_WRITABLE_PARTICULARS (TT, rv)
+
+#define DIRECT(TT,COUNT,ALLOC)  do                      \
+    {                                                   \
+      PREP (TT, COUNT);                                 \
+      memcpy (ST (rv, elt), src->pixels, ALLOC);        \
+      LATER (rv);                                       \
+    }                                                   \
+  while (0)
+
+#define SNARF(TT)  DIRECT (TT, len, sz)
+
+  /* Do it!  */
+  if (BOUNDP (squash) && NOT_FALSEP (squash))
+    DIRECT (u8, sz, sz);
+  else
+    switch (bypp)
+      {
+      case 4: SNARF (u32); break;
+      case 2: SNARF (u16); break;
+      case 1: SNARF (u8);  break;
+      case 3:
+        {
+          size_t i, lsbidx = SDL_LIL_ENDIAN == SDL_BYTEORDER ? 0 : 2;
+          const uint8_t *rp;
+          uint32_t *wp;
+
+          PREP (u32, len);
+          for (i = 0, rp = src->pixels, wp = ST (rv, elt);
+               i < len;
+               i++, rp += bypp, wp++)
+            *wp = ((rp[2 - lsbidx] << 16)
+                   | (rp[1]        << 8)
+                   | (rp[lsbidx]   << 0));
+          LATER (rv);
+        }
+      }
+
+  return rv;
+
+#undef SNARF
+#undef DIRECT
+#undef PREP
+#undef FUNC_NAME
+}
+
+
 
+#include "k/alphalim.c"
+#include "k/alphalimold.c"
+
 void
 gsdl_init_surface (void)
 {
@@ -614,7 +778,14 @@ gsdl_init_surface (void)
 #include "sdlsurface.x"
 
   /* alpha constants */
-  alpha_enums = DEFINE_ENUM ("alpha-enums", alpha_eback);
+  {
+    kp_init_t allp[] = {
+      { &alpha_enums, &alphalimold_kp },
+      { &kp_alpha, &alphalim_kp }
+    };
+
+    REGISTER_KP_V (allp);
+  }
 }
 
 /* sdlsurface.c ends here */
